from __future__ import absolute_import

from x3dh.exceptions import KeyExchangeException

from .exceptions import UnknownKeyException
from .state import make as make_State

import base64
import copy
import time

def make(backend):
    class X3DHDoubleRatchet(make_State(backend)):
        def __init__(self):
            super(X3DHDoubleRatchet, self).__init__()

            self.__bound_otpks = {}
            self.__pre_key_messages = {}

        def serialize(self):
            bound_otpks = {}

            for bare_jid in self.__bound_otpks:
                bound_otpks[bare_jid] = {}

                for device in self.__bound_otpks[bare_jid]:
                    otpk = self.__bound_otpks[bare_jid][device]

                    bound_otpks[bare_jid][device] = {
                        "otpk" : base64.b64encode(otpk["otpk"]).decode("US-ASCII"),
                        "id"   : otpk["id"]
                    }

            pk_messages = {}

            for otpk, value in self.__pre_key_messages.items():
                otpk = base64.b64encode(otpk).decode("US-ASCII")
                pk_messages[otpk] = copy.deepcopy(value)

            return {
                "super"       : super(X3DHDoubleRatchet, self).serialize(),
                "bound_otpks" : bound_otpks,
                "pk_messages" : pk_messages
            }

        @classmethod
        def fromSerialized(cls, serialized, *args, **kwargs):
            self = super(X3DHDoubleRatchet, cls).fromSerialized(
                serialized["super"],
                *args,
                **kwargs
            )

            bound_otpks = {}

            for bare_jid in serialized["bound_otpks"]:
                bound_otpks[bare_jid] = {}

                for device in serialized["bound_otpks"][bare_jid]:
                    otpk = serialized["bound_otpks"][bare_jid][device]

                    bound_otpks[bare_jid][device] = {
                        "otpk" : base64.b64decode(otpk["otpk"].encode("US-ASCII")),
                        "id"   : otpk["id"]
                    }

            pk_messages = {}

            for otpk, value in serialized["pk_messages"].items():
                otpk = base64.b64decode(otpk.encode("US-ASCII"))
                pk_messages[otpk] = copy.deepcopy(value)

            self.__bound_otpks = bound_otpks
            self.__pre_key_messages = pk_messages

            return self

        def getSharedSecretActive(
            self,
            other_public_bundle,
            _DEBUG_sendingRatchetKey = None,
            *args,
            **kwargs
        ):
            session_init_data = super(X3DHDoubleRatchet, self).getSharedSecretActive(
                other_public_bundle,
                *args,
                **kwargs
            )

            own_key = None
            if _DEBUG_sendingRatchetKey != None:
                import logging

                logging.getLogger("omemo.X3DHDoubleRatchet").error(
                    "WARNING: RUNNING UNSAFE DEBUG-ONLY OPERATION"
                )

                own_key = _DEBUG_sendingRatchetKey

            # When actively initializing a session
            # - The shared secret becomes the root key
            # - The public SPK used for X3DH becomes the other's enc for the dh ratchet
            # - The associated data calculated by X3DH becomes the ad used by the
            #   double ratchet encryption/decryption
            session_init_data["dr"] = backend.DoubleRatchet(
                session_init_data["ad"],
                session_init_data["sk"],
                own_key = own_key,
                other_pub = other_public_bundle.spk["key"]
            )

            # The shared secret and ad values are now irrelevant
            del session_init_data["sk"]
            del session_init_data["ad"]

            self.__compressSessionInitData(session_init_data, other_public_bundle)

            return session_init_data

        def getSharedSecretPassive(
            self,
            session_init_data,
            bare_jid,
            device,
            otpk_policy,
            additional_information = None
        ):
            self.__decompressSessionInitData(session_init_data, bare_jid, device)

            self.__preKeyMessageReceived(
                session_init_data["otpk"],
                additional_information
            )

            session_data = super(X3DHDoubleRatchet, self).getSharedSecretPassive(
                session_init_data,
                keep_otpk = True
            )

            # Decide whether to keep this OTPK
            self.__decideBoundOTPK(bare_jid, device, otpk_policy)

            # When passively initializing the session
            # - The shared secret becomes the root key
            # - The public SPK used by the active part for X3DH becomes the own dh ratchet
            #   key
            # - The associated data calculated by X3DH becomes the ad used by the double
            #   ratchet encryption/decryption
            return backend.DoubleRatchet(
                session_data["ad"],
                session_data["sk"],
                own_key = self.spk
            )

        def __compressSessionInitData(self, session_init_data, bundle):
            """
            Compress the session initialization data by replacing keys with their ids.
            """

            to_other = session_init_data["to_other"]

            to_other["otpk_id"] = bundle.findOTPKId(to_other["otpk"])
            to_other["spk_id"]  = bundle.findSPKId(to_other["spk"])

            del to_other["otpk"]
            del to_other["spk"]

        def __decompressSessionInitData(self, session_init_data, bare_jid, device):
            """
            Decompress the session initialization data by replacing key ids with the keys.
            """

            session_init_data["spk"] = self.getSPK(session_init_data["spk_id"])
            del session_init_data["spk_id"]

            otpk_id = self.getBoundOTPKId(bare_jid, device)

            # Check, whether the bare_jid+device combination is already bound to some OTPK
            if otpk_id:
                # If it is, check whether the OTPK ids match
                if otpk_id == session_init_data["otpk_id"]:
                    session_init_data["otpk"] = self.getBoundOTPK(bare_jid, device)

                # If the OTPK ids don't match, consider the old bound OTPK as deleteable
                # and bind the new OTPK
                else:
                    self.deleteBoundOTPK(bare_jid, device)

                    session_init_data["otpk"] = self.__bindOTPK(
                        bare_jid,
                        device,
                        session_init_data["otpk_id"]
                    )
            else:
                # If it is not, get the OTPK from the id and bind the bare_jid+device
                # combination to it
                session_init_data["otpk"] = self.__bindOTPK(
                    bare_jid,
                    device,
                    session_init_data["otpk_id"]
                )

            del session_init_data["otpk_id"]

        def __preKeyMessageReceived(self, otpk, additional_information = None):
            # Add an entry to the received PreKeyMessage data
            self.__pre_key_messages[otpk] = self.__pre_key_messages.get(otpk, [])
            self.__pre_key_messages[otpk].append({
                "timestamp": time.time(),
                "answers":   [],
                "additional_information": additional_information,
            })

        def getBoundOTPK(self, bare_jid, device):
            try:
                return self.__bound_otpks[bare_jid][device]["otpk"]
            except KeyError:
                return None

        def getBoundOTPKId(self, bare_jid, device):
            try:
                return self.__bound_otpks[bare_jid][device]["id"]
            except KeyError:
                return None

        def hasBoundOTPK(self, bare_jid, device):
            return True if self.getBoundOTPK(bare_jid, device) else False

        def respondedTo(self, bare_jid, device):
            bound_otpk_id = self.getBoundOTPK(bare_jid, device)

            self.__pre_key_messages[bound_otpk_id][-1]["answers"].append(time.time())

        def __decideBoundOTPK(self, bare_jid, device, otpk_policy):
            bound_otpk_id = self.getBoundOTPK(bare_jid, device)

            if not otpk_policy.decideOTPK(self.__pre_key_messages[bound_otpk_id]):
                self.deleteBoundOTPK(bare_jid, device)

        def deleteBoundOTPK(self, bare_jid, device):
            otpk = self.getBoundOTPK(bare_jid, device)

            if otpk:
                del self.__pre_key_messages[otpk]
                del self.__bound_otpks[bare_jid][device]
                self.deleteOTPK(otpk)

        def __bindOTPK(self, bare_jid, device, otpk_id):
            try:
                otpk = self.getOTPK(otpk_id)
            except UnknownKeyException:
                raise KeyExchangeException(
                    "The OTPK used for this session initialization has been deleted, " +
                    "the session can not be initiated."
                )

            self.__bound_otpks[bare_jid] = self.__bound_otpks.get(bare_jid, {})
            self.__bound_otpks[bare_jid][device] = {
                "otpk": otpk,
                "id": otpk_id
            }

            self.__pre_key_messages[otpk] = []

            self.hideFromPublicBundle(otpk)

            return otpk

    return X3DHDoubleRatchet



from __future__ import absolute_import

import copy
import logging
import os

from cryptography.hazmat.primitives.ciphers.aead import AESGCM

from x3dh.exceptions import KeyExchangeException

from . import promise
from . import storagewrapper
from .exceptions import *
from .x3dhdoubleratchet import make as make_X3DHDoubleRatchet

def d(*args, **kwargs):
    logging.getLogger("omemo.SessionManager").debug(*args, **kwargs)

def w(*args, **kwargs):
    logging.getLogger("omemo.SessionManager").warning(*args, **kwargs)

def e(*args, **kwargs):
    logging.getLogger("omemo.SessionManager").error(*args, **kwargs)

def checkSelf(self, *args, **kwargs):
    return self._storage.is_async

def checkPositionalArgument(position):
    def _checkPositionalArgument(*args, **kwargs):
        return args[position].is_async

    return _checkPositionalArgument

def checkConst(const):
    def _checkConst(*args, **kwargs):
        return const

    return _checkConst

class SessionManager(object):
    @classmethod
    @promise.maybe_coroutine(checkPositionalArgument(1))
    def create(cls, storage, otpk_policy, backend, my_bare_jid, my_device_id):
        self = cls()

        self._storage = storagewrapper.StorageWrapper(storage)

        self.__otpk_policy = otpk_policy

        self.__state = None

        self.__my_bare_jid  = my_bare_jid
        self.__my_device_id = my_device_id

        self.__devices_cache  = {}
        self.__sessions_cache = {}

        self.__backend = backend
        self.__X3DHDoubleRatchet = make_X3DHDoubleRatchet(self.__backend)

        yield self.__prepare()

        promise.returnValue(self)

    @promise.maybe_coroutine(checkSelf)
    def __prepare(self):
        state = yield self._storage.loadState()
        if state == None:
            self.__state = self.__X3DHDoubleRatchet()

            yield self._storage.storeState(self.__state.serialize())
            yield self._storage.storeActiveDevices(self.__my_bare_jid, [
                self.__my_device_id
            ])
        else:
            self.__state = self.__X3DHDoubleRatchet.fromSerialized(state)

        own_data = yield self._storage.loadOwnData()
        if own_data == None:

            yield self._storage.storeOwnData(self.__my_bare_jid, self.__my_device_id)
        else:
            if (not self.__my_bare_jid  == own_data["own_bare_jid"] or
                not self.__my_device_id == own_data["own_device_id"]):
                raise InconsistentInfoException(
                    "Given storage is only usable for jid \"" + own_data["own_bare_jid"] +
                    "\" on device " + str(own_data["own_device_id"]) + "."
                )

    @promise.maybe_coroutine(checkSelf)
    def __listDevices(self, bare_jid):
        if not (bare_jid in self.__devices_cache):
            active   = yield self._storage.loadActiveDevices(bare_jid)
            inactive = yield self._storage.loadInactiveDevices(bare_jid)

            self.__devices_cache[bare_jid] = {
                "active"   : set(active),
                "inactive" : set(inactive)
            }

        promise.returnValue(copy.deepcopy(self.__devices_cache[bare_jid]))

    @promise.maybe_coroutine(checkSelf)
    def __loadSession(self, bare_jid, device):
        self.__sessions_cache[bare_jid] = self.__sessions_cache.get(bare_jid, {})

        if not (device in self.__sessions_cache[bare_jid]):
            session = yield self._storage.loadSession(bare_jid, device)

            if session != None:
                session = self.__backend.DoubleRatchet.fromSerialized(session)

            self.__sessions_cache[bare_jid][device] = session

        promise.returnValue(self.__sessions_cache[bare_jid][device])

    @promise.maybe_coroutine(checkSelf)
    def __storeSession(self, bare_jid, device, session):
        self.__sessions_cache[bare_jid] = self.__sessions_cache.get(bare_jid, {})
        self.__sessions_cache[bare_jid][device] = session

        yield self._storage.storeSession(bare_jid, device, session.serialize())

    def __loggingEncryptMessageCallback(self, e, bare_jid, device):
        w(
            "Exception during encryption for device " +
            str(device) +
            " of bare jid " +
            bare_jid +
            ": " +
            str(e.__class__.__name__),
            exc_info = e
        )

    @promise.maybe_coroutine(checkSelf)
    def __encryptMessage(
        self,
        bare_jids,
        plaintext,
        bundles = None,
        devices = None,
        callback = None,
        always_trust = False,
        dry_run = False,
        _DEBUG_ek = None,
        _DEBUG_sendingRatchetKey = None
    ):
        # Lift a single bare_jid into a list
        if not isinstance(bare_jids, list):
            bare_jids = [ bare_jids ]

        # Add the own bare_jid to the list
        bare_jids = set(bare_jids) | set([ self.__my_bare_jid ])

        # If no callback was passed, log the exceptions
        if not callback:
            callback = self.__loggingEncryptMessageCallback

        # If no bundles were passed, default to an empty dict
        if bundles == None:
            bundles = {}

        if devices != None:
            devices = { bare_jid: devices.get(bare_jid, []) for bare_jid in bare_jids }
        else:
            devices = {}

            for bare_jid in bare_jids:
                # Load all active devices for this bare_jid
                devices[bare_jid] = (yield self.__listDevices(bare_jid))["active"]

                # If there are no active devices for this jid, generate an exception
                if len(devices[bare_jid]) == 0:
                    del devices[bare_jid]
                    callback(NoDevicesException(), bare_jid, None)

        # Don't encrypt the message for the sending device
        try:
            devices[self.__my_bare_jid].remove(self.__my_device_id)
        except (KeyError, ValueError):
            pass

        # Store all encrypted messages into this array
        # The elements will look like this:
        # {
        #     "rid"      : receiver_id  :int,
        #     "pre_key"  : pre_key      :bool,
        #     "message"  : message_data :bytes,
        #     "bare_jid" : bare_jid     :string
        # }
        messages = []

        aes_gcm_key = AESGCM.generate_key(bit_length = 128)
        aes_gcm_iv  = os.urandom(16)

        aes_gcm = AESGCM(aes_gcm_key)

        ciphertext = aes_gcm.encrypt(aes_gcm_iv, plaintext, None)

        aes_gcm_tag = ciphertext[-16:]
        ciphertext  = ciphertext[:-16]

        @promise.maybe_coroutine(checkConst(checkSelf(self)))
        def __encryptAll(devices, bare_jid):
            encrypted_count = 0

            for device in devices:
                is_trusted = yield self._storage.isTrusted(bare_jid, device)
                if not is_trusted and not always_trust:
                    callback(UntrustedException(), bare_jid, device)
                    continue

                if self.__state.hasBoundOTPK(bare_jid, device):
                    if not dry_run:
                        self.__state.respondedTo(bare_jid, device)
                        yield self._storage.storeState(self.__state.serialize())

                dr = yield self.__loadSession(bare_jid, device)

                pre_key = dr == None

                if pre_key:
                    try:
                        bundle = bundles[bare_jid][device]
                    except KeyError:
                        callback(MissingBundleException(), bare_jid, device)
                        continue

                    if not dry_run:
                        try:
                            session_init_data = self.__state.getSharedSecretActive(
                                bundle,
                                _DEBUG_ek = _DEBUG_ek,
                                _DEBUG_sendingRatchetKey = _DEBUG_sendingRatchetKey
                            )
                        except KeyExchangeException as e:
                            callback(e, bare_jid, device)
                            continue

                        # Store the changed state
                        yield self._storage.storeState(self.__state.serialize())

                        dr                = session_init_data["dr"]
                        session_init_data = session_init_data["to_other"]

                        pre_key = True

                if not dry_run:
                    message = dr.encryptMessage(aes_gcm_key + aes_gcm_tag)

                    # Store the new/changed session
                    yield self.__storeSession(bare_jid, device, dr)

                    message_data = self.__backend.WireFormat.messageToWire(
                        message["ciphertext"],
                        message["header"],
                        { "DoubleRatchet": message["additional"] }
                    )

                    if pre_key:
                        message_data = self.__backend.WireFormat.preKeyMessageToWire(
                            session_init_data,
                            message_data,
                            { "DoubleRatchet": message["additional"] }
                        )

                    messages.append({
                        "message"  : message_data,
                        "pre_key"  : pre_key,
                        "bare_jid" : bare_jid,
                        "rid"      : device
                    })

                encrypted_count += 1

            if encrypted_count == 0:
                if bare_jid != self.__my_bare_jid:
                    callback(NoEligibleDevicesException(), bare_jid, None)

        for bare_jid, deviceList in devices.items():
            yield __encryptAll(deviceList, bare_jid)

        promise.returnValue({
            "iv": aes_gcm_iv,
            "sid": self.__my_device_id,
            "messages": messages,
            "payload": ciphertext,
            "cipher": aes_gcm
        })

    @promise.maybe_coroutine(checkSelf)
    def encryptMessage(self, *args, **kwargs):
        result = yield self.__encryptMessage(*args, **kwargs)
        del result["cipher"]
        promise.returnValue(result)

    @promise.maybe_coroutine(checkSelf)
    def encryptKeyTransportMessage(self, bare_jids, *args, **kwargs):
        result = yield self.__encryptMessage(bare_jids, b"", *args, **kwargs)
        del result["payload"]
        promise.returnValue(result)

    @promise.maybe_coroutine(checkSelf)
    def buildSession(
        self,
        bare_jid,
        device,
        bundle,
        callback = None,
        dry_run = False,
        _DEBUG_ek = None,
        _DEBUG_sendingRatchetKey = None
    ):
        """
        Special version of encryptKeyTransportMessage, which does not encrypt a
        new KeyTransportMessage for all devices of the receiver and all devices
        of the sender but encrypts it for just the one specific device of the
        receiver.

        This can be used to build a session with a specific device without
        sending an initial text message.
        """

        promise.returnValue((yield self.encryptKeyTransportMessage(
            bare_jid,
            { bare_jid: { device: bundle } },
            { bare_jid: [ device ] },
            callback,
            always_trust = True,
            dry_run = dry_run,
            _DEBUG_ek = _DEBUG_ek,
            _DEBUG_sendingRatchetKey = _DEBUG_sendingRatchetKey
        )))

    @promise.maybe_coroutine(checkSelf)
    def _decryptMessage(
        self,
        bare_jid,
        device,
        message,
        is_pre_key_message,
        additional_information = None,
        _DEBUG_newRatchetKey = None
    ):
        if is_pre_key_message:
            # Unpack the pre key message data
            message_and_init_data = self.__backend.WireFormat.preKeyMessageFromWire(
                message
            )

            # Prepare the DoubleRatchet
            dr = self.__state.getSharedSecretPassive(
                message_and_init_data["session_init_data"],
                bare_jid,
                device,
                self.__otpk_policy,
                additional_information
            )

            # Store the changed state
            yield self._storage.storeState(self.__state.serialize())

            # Store the new session
            yield self.__storeSession(bare_jid, device, dr)

            # Unpack the "normal" message that was wrapped into the PreKeyMessage
            message = message_and_init_data["message"]
        else:
            # If this is not part of a PreKeyMessage,
            # we received a normal Message and can safely delete the OTPK
            self.__state.deleteBoundOTPK(bare_jid, device)
            yield self._storage.storeState(self.__state.serialize())

        # Unpack the message data
        message_data = self.__backend.WireFormat.messageFromWire(message)

        # Load the session
        dr = yield self.__loadSession(bare_jid, device)

        if dr == None:
            raise NoSessionException(
                "Don't have a session with \"" + bare_jid + "\" on device " +
                str(device) + "."
            )

        # Get the concatenation of the AES GCM key and tag
        plaintext = dr.decryptMessage(
            message_data["ciphertext"],
            message_data["header"],
            _DEBUG_newRatchetKey = _DEBUG_newRatchetKey
        )

        # Check the authentication
        self.__backend.WireFormat.finalizeMessageFromWire(
            message,
            {
                "WireFormat": message_data["additional"],
                "DoubleRatchet": plaintext["additional"]
            }
        )

        # Store the changed session
        yield self.__storeSession(bare_jid, device, dr)

        promise.returnValue(plaintext["plaintext"])

    @promise.maybe_coroutine(checkSelf)
    def decryptMessage(
        self,
        bare_jid,
        device,
        iv,
        message,
        is_pre_key_message,
        payload = None,
        additional_information = None
    ):
        plaintext = yield self._decryptMessage(
            bare_jid,
            device,
            message,
            is_pre_key_message,
            additional_information
        )

        aes_gcm_key = plaintext[:16]
        aes_gcm_tag = plaintext[16:]

        aes_gcm = AESGCM(aes_gcm_key)

        if payload == None:
            # Return the prepared cipher
            promise.returnValue(( aes_gcm, None ))
        else:
            # Return the plaintext
            promise.returnValue((
                None,
                aes_gcm.decrypt(iv, payload + aes_gcm_tag, None)
            ))

    @promise.maybe_coroutine(checkSelf)
    def newDeviceList(self, devices, bare_jid):
        devices = set(devices)

        if bare_jid == self.__my_bare_jid:
            # The own device can never become inactive
            devices |= set([ self.__my_device_id ])

        devices_old = yield self.__listDevices(bare_jid)
        devices_old = devices_old["active"] | devices_old["inactive"]

        active   = devices
        inactive = devices_old - active

        self.__devices_cache[bare_jid] = { "active": active, "inactive": inactive }

        yield self._storage.storeActiveDevices(bare_jid, active)
        yield self._storage.storeInactiveDevices(bare_jid, inactive)

    @promise.maybe_coroutine(checkSelf)
    def getDevices(self, bare_jid = None):
        if not bare_jid:
            bare_jid = self.__my_bare_jid

        result = yield self.__listDevices(bare_jid)
        promise.returnValue(result)

    @property
    def public_bundle(self):
        return self.__state.getPublicBundle()

    @property
    def fingerprint(self):
        return self.public_bundle.fingerprint

    @property
    def republish_bundle(self):
        return self.__state.changed

    ###############################
    # DEBUG ADDITIONS, DO NOT USE #
    ###############################

    def _DEBUG_simulatePreKeyMessage(
        self,
        other_session_manager,
        otpk_id,
        ek,
        sending_ratchet_key
    ):
        from .extendedpublicbundle import ExtendedPublicBundle

        e("WARNING: RUNNING UNSAFE DEBUG-ONLY OPERATION")
        d("Simulating pre key message.")

        other_bundle = other_session_manager.state.getPublicBundle()

        ik  = other_bundle.ik
        spk = other_bundle.spk
        spk_signature = other_bundle.spk_signature
        otpks = [ otpk for otpk in other_bundle.otpks if otpk["id"] == otpk_id ]

        other_bundle = ExtendedPublicBundle(ik, spk, spk_signature, otpks)

        my_own_data    = self._storage.loadOwnData()
        other_own_data = other_session_manager._storage.loadOwnData()

        self.newDeviceList(
            [ my_own_data["own_device_id"] ],
            my_own_data["own_bare_jid"]
        )

        self.newDeviceList(
            [ other_own_data["own_device_id"] ],
            other_own_data["own_bare_jid"]
        )

        self.buildSession(
            other_own_data["own_bare_jid"],
            other_own_data["own_device_id"],
            other_bundle,
            _DEBUG_ek = ek,
            _DEBUG_sendingRatchetKey = sending_ratchet_key
        )

    def _DEBUG_simulateMessage(self, other_session_manager):
        e("WARNING: RUNNING UNSAFE DEBUG-ONLY OPERATION")
        d("Simulating message.")

        my_own_data    = self._storage.loadOwnData()
        other_own_data = other_session_manager._storage.loadOwnData()

        self.newDeviceList(
            [ my_own_data["own_device_id"] ],
            my_own_data["own_bare_jid"]
        )

        self.newDeviceList(
            [ other_own_data["own_device_id"] ],
            other_own_data["own_bare_jid"]
        )

        self.encryptKeyTransportMessage(other_own_data["own_bare_jid"])

    def _DEBUG_compareState(self, other_session_manager, state):
        e("WARNING: RUNNING UNSAFE DEBUG-ONLY OPERATION")
        d("Comparing states.")

        import base64

        other_own_data = other_session_manager._storage.loadOwnData()

        dr = self.__loadSession(
            other_own_data["own_bare_jid"],
            other_own_data["own_device_id"]
        )

        assert dr != None

        dr = dr.serialize()

        root_key = base64.b64decode(dr["root_chain"]["super"]["key"].encode("US-ASCII"))
        assert root_key == state["rootKey"]

        try:
            schain_key = dr["skr"]["super"]["schain"]["super"]["super"]["key"]
            schain_key = base64.b64decode(schain_key.encode("US-ASCII"))
        except TypeError:
            pass
        else:
            assert schain_key == state["senderChainKey"]

        if state["receiverChainKey"] != None:
            try:
                rchain_key = dr["skr"]["super"]["rchain"]["super"]["super"]["key"]
                rchain_key = base64.b64decode(rchain_key.encode("US-ASCII"))
            except TypeError:
                pass
            else:
                assert rchain_key == state["receiverChainKey"]



class Backend(object):
    def __init__(self, WireFormat, X3DHState, DoubleRatchet):
        self.__WireFormat    = WireFormat
        self.__X3DHState     = X3DHState
        self.__DoubleRatchet = DoubleRatchet

    @property
    def WireFormat(self):
        return self.__WireFormat

    @property
    def X3DHState(self):
        return self.__X3DHState

    @property
    def DoubleRatchet(self):
        return self.__DoubleRatchet



from __future__ import absolute_import

import functools
import types

class ReturnValueException(Exception):
    def __init__(self, value):
        self.__value = value

    @property
    def value(self):
        return self.__value

class RejectedException(Exception):
    def __init__(self, reason):
        self.__reason = reason
    
    def __eq__(self, other):
        if isinstance(other, RejectedException):
            return self.__reason == other.reason

        return False

    def __hash__(self):
        return hash(self.__reason)

    @property
    def reason(self):
        return self.__reason

class InvalidCoroutineException(Exception):
    def __init__(self, reason):
        self.__reason = reason
    
    def __eq__(self, other):
        if isinstance(other, InvalidCoroutineException):
            return self.__reason == other.reason

        return False

    def __hash__(self):
        return hash(self.__reason)

    @property
    def reason(self):
        return self.__reason

class Promise(object):
    PENDING   = "PENDING"
    FULFILLED = "FULFILLED"
    REJECTED  = "REJECTED"

    def __init__(self, code):
        self.__state  = Promise.PENDING
        self.__value  = None
        self.__reason = None

        self.__onfulfilled = []
        self.__onrejected  = []

        try:
            code(self.__resolve, self.__reject)
        except BaseException as e:
            self.__reject(e)

    def __resolve(self, value):
        if self.__state == Promise.PENDING:
            self.__value = value
            self.__state = Promise.FULFILLED

            while len(self.__onfulfilled) > 0:
                listener = self.__onfulfilled.pop(0)
                listener(self.__value)

    def __reject(self, reason):
        if self.__state == Promise.PENDING:
            self.__reason = reason
            self.__state  = Promise.REJECTED

            while len(self.__onrejected) > 0:
                listener = self.__onrejected.pop(0)
                listener(self.__reason)

    def then(self, onfulfilled, onrejected):
        if callable(onfulfilled):
            if self.__state == Promise.PENDING:
                self.__onfulfilled.append(onfulfilled)
            
            if self.__state == Promise.FULFILLED:
                onfulfilled(self.__value)

        if callable(onrejected):
            if self.__state == Promise.PENDING:
                self.__onrejected.append(onrejected)
            
            if self.__state == Promise.REJECTED:
                onrejected(self.__reason)

    @property
    def done(self):
        return not self.__state == Promise.PENDING

    @property
    def fulfilled(self):
        return self.__state == Promise.FULFILLED

    @property
    def rejected(self):
        return self.__state == Promise.REJECTED

    def __str__(self):
        if self.__state == Promise.PENDING:
            return "Pending Promise object"

        if self.__state == Promise.FULFILLED:
            return "Fulfilled Promise object with value: " + str(self.__value)

        if self.__state == Promise.REJECTED:
            return "Rejected Promise object with reason: " + str(self.__reason)

    @property
    def state(self):
        return self.__state

    @property
    def value(self):
        return self.__value

    @property
    def reason(self):
        return self.__reason

    @classmethod
    def resolve(cls, value):
        return cls(lambda resolve, reject: resolve(value))

    @classmethod
    def reject(cls, reason):
        return cls(lambda resolve, reject: reject(reason))

def returnValue(value):
    """
    In Python 2 we are not allowed to return from a generator function.

    Instead, we have to raise the StopIteration exceptions ourselves to return from the
    coroutine.

    Python 3.7 for some fucking reason changed it so you can't raise a StopIteration
    exception yourself. Really fucking great idea. Let's make it harder every version to
    write software for Python 2 and 3.

    For this awesome reason, we don't raise a StopIteration exception but a self-made
    exception called ReturnValueException.
    """

    raise ReturnValueException(value)

def coroutine(f):
    """
    Implementation of a coroutine.

    Use as a decorator:
    @coroutine
    def foo():
        result = yield somePromise

    The function passed should be a generator yielding instances of the Promise class
    (or compatible).
    The coroutine waits for the Promise to resolve and sends the result (or the error)
    back into the generator function.
    This simulates sequential execution which in reality can be asynchonous.
    """

    @functools.wraps(f)
    def _coroutine(*args, **kwargs):
        def _resolver(resolve, reject):
            try:
                generator = f(*args, **kwargs)
            except BaseException as e:
                # Special case for a function that throws immediately
                reject(e)
            else:
                # Special case for a function that returns immediately
                if not isinstance(generator, types.GeneratorType):
                    resolve(generator)
                else:
                    def _step(previous, previous_type):
                        element = None

                        try:
                            if previous_type == None:
                                element = next(generator)
                            elif previous_type:
                                element = generator.send(previous)
                            else:
                                if not isinstance(previous, BaseException):
                                    previous = RejectedException(previous)

                                element = generator.throw(previous)
                        except StopIteration as e:
                            resolve(getattr(e, "value", None))
                        except ReturnValueException as e:
                            resolve(e.value)
                        except BaseException as e:
                            reject(e)
                        else:
                            try:
                                element.then(
                                    lambda value  : _step(value, True),
                                    lambda reason : _step(reason, False)
                                )
                            except AttributeError:
                                reject(InvalidCoroutineException(element))

                    _step(None, None)

        return Promise(_resolver)

    return _coroutine

def no_coroutine(f):
    """
    This is not a coroutine ;)

    Use as a decorator:
    @no_coroutine
    def foo():
        five = yield 5
        print(yield "hello")

    The function passed should be a generator yielding whatever you feel like.
    The yielded values instantly get passed back into the generator.
    It's basically the same as if you didn't use yield at all.
    The example above is equivalent to:
    def foo():
        five = 5
        print("hello")

    Why?
    This is the counterpart to coroutine used by maybe_coroutine below.
    """

    @functools.wraps(f)
    def _no_coroutine(*args, **kwargs):
        generator = f(*args, **kwargs)

        # Special case for a function that returns immediately
        if not isinstance(generator, types.GeneratorType):
            return generator

        def _step(previous, first):
            element = None

            try:
                if first:
                    element = next(generator)
                else:
                    element = generator.send(previous)
            except StopIteration as e:
                return getattr(e, "value", None)
            except ReturnValueException as e:
                return e.value
            else:
                return _step(element, False)

        return _step(None, True)

    return _no_coroutine

def maybe_coroutine(decide):
    """
    Either be a coroutine or not.

    Use as a decorator:
    @maybe_coroutine(lambda maybeAPromise: return isinstance(maybeAPromise, Promise))
    def foo(maybeAPromise):
        result = yield maybeAPromise
        print("hello")
        return result

    The function passed should be a generator yielding either only Promises or whatever
    you feel like.
    The decide parameter must be a function which gets called with the same parameters as
    the function to decide whether this is a coroutine or not.
    Using this it is possible to either make the function a coroutine or not based on a
    parameter to the function call.
    Let's explain the example above:

    # If the maybeAPromise is an instance of Promise,
    # we want the foo function to act as a coroutine.
    # If the maybeAPromise is not an instance of Promise,
    # we want the foo function to act like any other normal synchronous function.
    @maybe_coroutine(lambda maybeAPromise: return isinstance(maybeAPromise, Promise))
    def foo(maybeAPromise):
        # If isinstance(maybeAPromise, Promise), foo behaves like a coroutine,
        # thus maybeAPromise will get resolved asynchronously and the result will be
        # pushed back here.
        # Otherwise, foo behaves like no_coroutine,
        # just pushing the exact value of maybeAPromise back into the generator.
        result = yield maybeAPromise
        print("hello")
        return result
    """

    def _maybe_coroutine(f):
        @functools.wraps(f)
        def __maybe_coroutine(*args, **kwargs):
            if decide(*args, **kwargs):
                return coroutine(f)(*args, **kwargs)
            else:
                return no_coroutine(f)(*args, **kwargs)
        return __maybe_coroutine
    return _maybe_coroutine



from __future__ import absolute_import

from .version import __version__

from . import backends
from . import promise
from . import util
from .extendedpublicbundle import ExtendedPublicBundle
from .otpkpolicy import OTPKPolicy
from .sessionmanager import SessionManager
from .state import make as make_State
from .storage import Storage
from .x3dhdoubleratchet import make as make_X3DHDoubleRatchet



from __future__ import absolute_import

import base64

import x3dh

from .exceptions import UnknownKeyException
from .extendedpublicbundle import ExtendedPublicBundle

def make(backend):
    class State(backend.X3DHState):
        def __init__(self):
            super(State, self).__init__()

            self.__spk_id  = 0
            self.__spk_pub = None

            self.__otpk_id_counter = 0
            self.__otpk_ids = {}

        def serialize(self):
            spk = self.__spk_pub
            spk = None if spk == None else base64.b64encode(spk).decode("US-ASCII")

            otpk_ids = {}

            for key, value in self.__otpk_ids.items():
                otpk_ids[base64.b64encode(key).decode("US-ASCII")] = value

            return {
                "super": super(State, self).serialize(),
                "spk_id": self.__spk_id,
                "spk_pub": spk,
                "otpk_id_counter": self.__otpk_id_counter,
                "otpk_ids": otpk_ids
            }

        @classmethod
        def fromSerialized(cls, serialized, *args, **kwargs):
            self = super(State, cls).fromSerialized(
                serialized["super"],
                *args,
                **kwargs
            )

            spk = serialized["spk_pub"]
            spk = None if spk == None else base64.b64decode(spk.encode("US-ASCII"))

            otpk_ids = {}

            for key, value in serialized["otpk_ids"].items():
                otpk_ids[base64.b64decode(key.encode("US-ASCII"))] = value

            self.__spk_id = serialized["spk_id"]
            self.__spk_pub = spk
            self.__otpk_id_counter = serialized["otpk_id_counter"]
            self.__otpk_ids = otpk_ids

            return self

        def getPublicBundle(self):
            """
            The current OMEMO standard works with ids instead of sending full public keys
            whenever possible, probably to reduce traffic.
            This is not part of the core specification though, so it has to be added here.
            It is added in the getPublicBundle method, because this method is the only way
            to get public data and is the perfect spot to update ids.
            """

            bundle = super(State, self).getPublicBundle()

            self.__updateIDs()

            return self.__extendBundle(bundle)

        def __updateIDs(self):
            # Check, whether the spk has changed and assign it the next id in that case
            if self.spk.pub != self.__spk_pub:
                self.__spk_pub = self.spk.pub
                self.__spk_id += 1

            otpks        = [ otpk.pub for otpk in self.otpks ]
            hidden_otpks = [ otpk.pub for otpk in self.hidden_otpks ]

            # Synchronize the list of OTPKs.
            # First, remove all entries in the current dict,
            # that were removed from the official list.
            for otpk in list(self.__otpk_ids):
                if not (otpk in otpks or otpk in hidden_otpks):
                    del self.__otpk_ids[otpk]

            # Second, add new OTPKs to the dict and assign them ids
            for otpk in otpks:
                if not otpk in self.__otpk_ids:
                    self.__otpk_id_counter += 1
                    self.__otpk_ids[otpk] = self.__otpk_id_counter

        def __extendBundle(self, bundle):
            """
            Extend the bundle, adding the ids of the respective keys to all entries.
            """

            ik = bundle.ik

            spk = {
                "key": bundle.spk,
                "id": self.getSPKID(bundle.spk)
            }

            spk_signature = bundle.spk_signature

            otpks = [ {
                "key": otpk,
                "id": self.getOTPKID(otpk)
            } for otpk in bundle.otpks ]

            return ExtendedPublicBundle(ik, spk, spk_signature, otpks)

        def __reduceBundle(self, bundle):
            """
            Reduce the bundle, removing all ids of the respective keys from all entries.
            """

            ik = bundle.ik

            spk = bundle.spk["key"]

            spk_signature = bundle.spk_signature

            otpks = [ otpk["key"] for otpk in bundle.otpks ]

            return x3dh.PublicBundle(ik, spk, spk_signature, otpks)

        def getSPKID(self, spk):
            self.__updateIDs()

            # If the requested spk is the most recent one...
            if self.__spk_pub == spk:
                # ...return the id
                return self.__spk_id

            raise UnknownKeyException("Tried to get the id of an unknown SPK.")

        def getSPK(self, spk_id):
            self.__updateIDs()

            # If the requested spk id is the one contained in this bundle...
            if self.__spk_id == spk_id:
                # ...return the key
                return self.__spk_pub

            raise UnknownKeyException("Tried to get the SPK for an unknown id.")

        def getOTPKID(self, otpk):
            self.__updateIDs()

            otpk_id = self.__otpk_ids.get(otpk)

            if otpk_id == None:
                raise UnknownKeyException("Tried to get the id of an unknown OTPK.")

            return otpk_id

        def getOTPK(self, otpk_id):
            self.__updateIDs()

            otpks = [ x[0] for x in self.__otpk_ids.items() if x[1] == otpk_id ]

            if len(otpks) != 1:
                raise UnknownKeyException("Tried to get the OTPK for an unknown id.")

            return otpks[0]

        def getSharedSecretActive(self, other_public_bundle, *args, **kwargs):
            other_public_bundle = self.__reduceBundle(other_public_bundle)

            return super(State, self).getSharedSecretActive(
                other_public_bundle,
                *args,
                **kwargs
            )

    return State



from __future__ import absolute_import

from .promise import Promise
from .storage import Storage

def makeCBPromise(function, *args, **kwargs):
    """
    Take a function that reports its result using a callback and return a Promise that
    listenes for this callback.

    The function must accept a callback as its first parameter.
    The callback must take two arguments:
    - success : True or False
    - result  : The result of the operation if success is True or the error otherwise.
    """

    def _resolver(resolve, reject):
        function(
            lambda success, result: resolve(result) if success else reject(result),
            *args,
            **kwargs
        )

    return Promise(_resolver)

class StorageWrapper(Storage):
    def __init__(self, wrapped):
        self.__wrapped = wrapped

    def loadOwnData(self, *args, **kwargs):
        if self.__wrapped.is_async:
            return makeCBPromise(self.__wrapped.loadOwnData, *args, **kwargs)
        else:
            return self.__wrapped.loadOwnData(None, *args, **kwargs)

    def storeOwnData(self, *args, **kwargs):
        if self.__wrapped.is_async:
            return makeCBPromise(self.__wrapped.storeOwnData, *args, **kwargs)
        else:
            return self.__wrapped.storeOwnData(None, *args, **kwargs)

    def loadState(self, *args, **kwargs):
        if self.__wrapped.is_async:
            return makeCBPromise(self.__wrapped.loadState, *args, **kwargs)
        else:
            return self.__wrapped.loadState(None, *args, **kwargs)

    def storeState(self, *args, **kwargs):
        if self.__wrapped.is_async:
            return makeCBPromise(self.__wrapped.storeState, *args, **kwargs)
        else:
            return self.__wrapped.storeState(None, *args, **kwargs)

    def loadSession(self, *args, **kwargs):
        if self.__wrapped.is_async:
            return makeCBPromise(self.__wrapped.loadSession, *args, **kwargs)
        else:
            return self.__wrapped.loadSession(None, *args, **kwargs)

    def storeSession(self, *args, **kwargs):
        if self.__wrapped.is_async:
            return makeCBPromise(self.__wrapped.storeSession, *args, **kwargs)
        else:
            return self.__wrapped.storeSession(None, *args, **kwargs)

    def loadActiveDevices(self, *args, **kwargs):
        if self.__wrapped.is_async:
            return makeCBPromise(self.__wrapped.loadActiveDevices, *args, **kwargs)
        else:
            return self.__wrapped.loadActiveDevices(None, *args, **kwargs)

    def loadInactiveDevices(self, *args, **kwargs):
        if self.__wrapped.is_async:
            return makeCBPromise(self.__wrapped.loadInactiveDevices, *args, **kwargs)
        else:
            return self.__wrapped.loadInactiveDevices(None, *args, **kwargs)

    def storeActiveDevices(self, *args, **kwargs):
        if self.__wrapped.is_async:
            return makeCBPromise(self.__wrapped.storeActiveDevices, *args, **kwargs)
        else:
            return self.__wrapped.storeActiveDevices(None, *args, **kwargs)

    def storeInactiveDevices(self, *args, **kwargs):
        if self.__wrapped.is_async:
            return makeCBPromise(self.__wrapped.storeInactiveDevices, *args, **kwargs)
        else:
            return self.__wrapped.storeInactiveDevices(None, *args, **kwargs)

    def isTrusted(self, *args, **kwargs):
        if self.__wrapped.is_async:
            return makeCBPromise(self.__wrapped.isTrusted, *args, **kwargs)
        else:
            return self.__wrapped.isTrusted(None, *args, **kwargs)

    @property
    def is_async(self):
        return self.__wrapped.is_async



class OTPKPolicy(object):
    @staticmethod
    def decideOTPK(preKeyMessages):
        """
        Use the data passed to this method to decide, whether to keep an OTPK or not.
        Return True to keep the OTPK and False to delete it.

        The preKeyMessages parameter is a list of dictionaries with following structure:
        {
            # The UNIX timestamp that PreKeyMessage was received on
            "timestamp": int,

            # A list of UNIX timestamps, for each Message that answered this PreKeyMessage
            "answers": list<int>,

            # This key can be used by implementations to store any sort of additional
            # information about the message, which can be used for more complex logic to
            # decide whether to keep the one-time pre key. One example that would make a
            # lot of sense is a flag, which indicates whether the message was retrieved
            # from some storage mechanism like mam. Messages retrieved from mam should
            # probably not trigger one-time pre key deletion, because there might be more
            # pre key messages waiting in the mam catch-up that use the same one-time pre
            # key.
            # The value of this key must consist of Python primitives like ints, floats,
            # strings, booleans, lists, dictionaries or None (basically everything
            # json-serializable).
            "additional_information": any
        }
        """

        raise NotImplementedError



"""
The interface used by the SessionManager to persist data between runs.

There are two possible ways to implement the Storage class: synchronous or asynchronous.

The mode is determined by the result of the is_async method.

If the implementation is asynchronous, the callback parameter is a function that takes
two arguments:
- success: True or False
- result: The result of the operation if success is True or the error if success is False

If the implementation is synchronous, the callback parameter is None.
"""

class Storage(object):
    def loadOwnData(self, callback):
        """
        Load the own data.

        Return a dictionary of following structure:
        {
            "own_bare_jid"  : string,
            "own_device_id" : int
        }

        or None, if no own data was stored previously.
        """

        raise NotImplementedError

    def storeOwnData(self, callback, own_bare_jid, own_device_id):
        """
        Store given own data, overwriting previously stored data.
        """

        raise NotImplementedError

    def loadState(self, callback):
        """
        Load the state.

        Return the stored structure or None, if no state was stored previously.
        """

        raise NotImplementedError

    def storeState(self, callback, state):
        """
        Store the state, overwriting the old state, if it exists.

        state is passed as a serializable object, that means it consist of a combination
        of the following types:
        - dictionaries
        - lists
        - strings
        - integers
        - floats
        - booleans
        - None

        You can dump this object using for example the json module.

        For more information on how the state object is structured, look at the
        omemo.X3DHDoubleRatchet.serialize method.
        """

        raise NotImplementedError

    def loadSession(self, callback, bare_jid, device_id):
        """
        Load a session with given bare_jid and device id.

        bare_jid is passed as a string, device_id as an integer.

        Return either the structure previously stored or None.
        """

        raise NotImplementedError

    def storeSession(self, callback, bare_jid, device_id, session):
        """
        Store a session for given bare_jid and device id, overwriting the previous
        session, if it exists.

        bare_jid is passed as string, device_id as an integer.

        session is passed as a serializable object, that means it consist of a combination
        of the following types:
        - dictionaries
        - lists
        - strings
        - integers
        - floats
        - booleans
        - None

        You can dump this object using for example the json module.

        For more information on how the session object is structured, look at the
        serialize method of the DoubleRatchet implementation offered by the backend you're
        running.
        """

        raise NotImplementedError

    def loadActiveDevices(self, callback, bare_jid):
        """
        Load the list of active devices for a given bare_jid.

        An "active device" is a device, which is listed in the most recent version of
        the device list pep node.

        bare_jid is passed as a string, the result is expected to be a list of integers.
        """

        raise NotImplementedError

    def loadInactiveDevices(self, callback, bare_jid):
        """
        Load the list of active devices for a given bare_jid.

        An "inactive device" is a device, which was listed in an older version of
        the device list pep node, but is NOT listed in the most recent version.

        bare_jid is passed as a string, the result is expected to be a list of integers.
        """

        raise NotImplementedError

    def storeActiveDevices(self, callback, bare_jid, devices):
        """
        Store the active devices for given bare_jid, overwriting the old stored list,
        if it exists.

        bare_jid is passed as a string, devices as a list of integers.
        """

        raise NotImplementedError

    def storeInactiveDevices(self, callback, bare_jid, devices):
        """
        Store the inactive devices for given bare_jid, overwriting the old stored list,
        if it exists.

        bare_jid is passed as a string, devices as a list of integers.
        """

        raise NotImplementedError

    def isTrusted(self, callback, bare_jid, device):
        """
        Return, whether the given device of given bare_jid is trusted.

        bare_jid is passed as a string, device as an integer.
        """

        raise NotImplementedError

    @property
    def is_async(self):
        """
        Return, whether this implementation is asynchronous.

        Read the introduction to this module above for details on what this value changes.
        """

        raise NotImplementedError



from __future__ import absolute_import

from .inconsistentinfoexception import InconsistentInfoException
from .missingbundleexception import MissingBundleException
from .nodevicesexception import NoDevicesException
from .noeligibledevicesexception import NoEligibleDevicesException
from .nosessionexception import NoSessionException
from .unknownkeyexception import UnknownKeyException
from .untrustedexception import UntrustedException
from .wireformatexception import WireFormatException



from __future__ import absolute_import

import os
import struct

DEVICE_ID_MIN = 1
DEVICE_ID_MAX = 2 ** 31 - 1

def generateDeviceID(blacklist = []):
    while True:
        device_id = struct.unpack(">L", os.urandom(4))[0]

        if device_id < DEVICE_ID_MIN or device_id > DEVICE_ID_MAX:
            continue

        if device_id in blacklist:
            continue

        return device_id



class WireFormat(object):
    @staticmethod
    def messageFromWire(obj):
        raise NotImplementedError

    @staticmethod
    def finalizeMessageFromWire(obj, additional):
        raise NotImplementedError

    @staticmethod
    def messageToWire(ciphertext, header, additional):
        raise NotImplementedError

    @staticmethod
    def preKeyMessageFromWire(obj):
        raise NotImplementedError

    @staticmethod
    def finalizePreKeyMessageFromWire(obj, additional):
        raise NotImplementedError

    @staticmethod
    def preKeyMessageToWire(session_init_data, message, additional):
        raise NotImplementedError



from __future__ import absolute_import

from .exceptions import UnknownKeyException

class ExtendedPublicBundle(object):
    """
    This class looks exactly the same as the PublicBundle class, but the types of the
    fields are a bit different:

    The spk field is not a key, but a dictionary containing the key and the id:
    spk = {
        "key" : key,
        "id"  : id
    }

    The otpks field is not an array of keys, but an array of dictionaries containing the
    key and the id:
    otpks = [
        {
            "key" : key,
            "id"  : id
        },
        {
            "key" : key,
            "id"  : id
        },
        ...
    ]
    """

    def __init__(self, ik, spk, spk_signature, otpks):
        self.__ik = ik
        self.__spk = spk
        self.__spk_signature = spk_signature
        self.__otpks = otpks

    @property
    def ik(self):
        return self.__ik

    @property
    def spk(self):
        return self.__spk

    @property
    def spk_signature(self):
        return self.__spk_signature

    @property
    def otpks(self):
        return self.__otpks

    def findOTPKId(self, otpk):
        otpks = list(filter(lambda x: x["key"] == otpk, self.otpks))

        if len(otpks) != 1:
            raise UnknownKeyException("Tried to get the id of an unknown OTPK.")

        return otpks[0]["id"]

    def findSPKId(self, spk):
        # If the requested spk is the one contained in this bundle...
        if self.spk["key"] == spk:
            # ...return the id
            return self.spk["id"]

        raise UnknownKeyException("Tried to get the id of an unknown SPK.")
